/* Copyright (C) 2002 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   This is a generic main loop for an application that's primarily an
   communications server.  It's primarily a just a small wrapper function
   for the magic select() system call.  
   
   $Log: main.c,v $
   Revision 1.3  2002/08/18 16:34:10  wb6ymh
   Added DumpMe() to cause core dumps on demand for debugging.
   Added call to DumpMe() in DeleteClient() when client being deleted isn't
   found in AVL tree. (caused infinite loop when EchoLink server when down
   recently).  I don't see the problem, the coredump will tell all.

   Revision 1.2  2002/08/12 17:10:47  wb6ymh
   Made TimeNow a global.
   Removed argument to GetTimeNow().

   Revision 1.1.1.1  2002/08/10 20:33:41  wb6ymh
   initial import

*/
#include "common.h"

#ifndef _WIN32
   // FreeBSD, Linux, etc..
   #include <stdio.h>
   #include <stdlib.h>
   #include <unistd.h>
   #include <netinet/in.h>
   #include <sys/socket.h>
   #ifdef TIME_WITH_SYS_TIME
      #include <sys/time.h>
      #include <time.h>
   #else
      #ifdef HAVE_SYS_TIME_H
         #include <sys/time.h>
      #else
         #include <time.h>
      #endif
   #endif
#ifdef HAVE_SYS_SIGNAL_H
   #include <sys/signal.h>
#endif
   #include <string.h>

#ifdef HAVE_LIMITS_H
   #include <limits.h>
#else
   #define  INT_MAX     0x7fffffff  /* assume 32 bit ints */
#endif

   #include <errno.h>
   #include <syslog.h>
   #include <stdarg.h>
   #ifdef HAVE_SYS_TIMEB_H
      #include <sys/timeb.h>
   #endif
#else
   // Windoze
   #include <stdio.h>
   #include <time.h>
   #include <sys\timeb.h>
   #include <io.h>
   #include <winsock2.h>
   #include <limits.h>
   #include <stdarg.h>
#endif

#include "avl.h"
#include "main.h"

#define MAX_SEND_SIZE   1024

int SN;
int DebugLevel;
int bShutdown = FALSE;
int bRunning = TRUE;
int bRefreshConfig = FALSE;
int bDaemon = FALSE;

time_t BootTime;
struct timeval TimeNow;

struct avl_table *ClientTree;

ClientInfo *pClientDebug = NULL;

void CallCurrentState(ClientInfo *pClient);

#ifndef _WIN32
// FreeBSD, Linux, etc..
void SigHandler(int Signal);
#else
BOOL SigHandler(DWORD Type);
#endif

int MainLoop()
{
   fd_set ReadFDset;
   fd_set WriteFDset;
   fd_set ExceptFDset;
   int Selret;
   ClientInfo *pClient;
   ClientInfo *pClient1;
   struct avl_traverser avl_trans;
   struct timeval SelTO;
   long t;
   int ReadWaitCnt;
   int WriteWaitCnt;
   int Err = 0;
   long TimeOutSeconds;

   BootTime = time(NULL);

#ifndef _WIN32
   signal(SIGHUP,SigHandler);
   signal(SIGTERM,SigHandler);
   signal(SIGINT,SigHandler);
#else
   SetConsoleCtrlHandler((PHANDLER_ROUTINE) SigHandler,TRUE);                           // add to list 
#endif

   while(bRunning) {
      if(bRefreshConfig) {
         bRefreshConfig = FALSE;
         if(ReadConfig()) {
            LOG_ERROR(("parse error reloading configuration file, exiting"));
            Err = ERR_CONFIG_FILE;
            break;
         }
      }

      if(bShutdown) {
         bShutdown = FALSE;
         Shutdown();
      }

      FD_ZERO(&ReadFDset);
      FD_ZERO(&WriteFDset);
      FD_ZERO(&ExceptFDset);

#ifndef _WIN32
      SelTO.tv_sec = INT_MAX;
#else
   // No infinite wait in Windoze even if no clients request a timeout
      SelTO.tv_sec = INT_MAX - 1;
#endif

      SelTO.tv_usec = 0;
      ReadWaitCnt = WriteWaitCnt = 0;

      pClient = (ClientInfo *) avl_t_first(&avl_trans,ClientTree);
      while(pClient != NULL) {
         pClient->bReadReady = 
         pClient->bWriteReady = 
         pClient->bTimeOut =
         pClient->bNew = FALSE;

         if(pClient->bReadWait) {
            ReadWaitCnt++;
            FD_SET(pClient->Socket,&ReadFDset);
         }

         if(pClient->bWriteWait) {
            WriteWaitCnt++;
            FD_SET(pClient->Socket,&WriteFDset);
         }

         if(pClient->bTimeWait) {
         // Set SelTO to the shortest timeout requested
            t = SelTO.tv_sec - pClient->Timeout.tv_sec;
            if(t == 0) {
               t = SelTO.tv_usec - pClient->Timeout.tv_usec;
            }
            
            if(t > 0) {
               SelTO = pClient->Timeout;
            }
         }
         pClient = (ClientInfo *) avl_t_next(&avl_trans);
      }

      if(SelTO.tv_sec != INT_MAX) {
      // Calculate delta T to shortest timeout requested
         GetTimeNow();
         SelTO.tv_usec -= TimeNow.tv_usec;
         SelTO.tv_sec -= TimeNow.tv_sec;
         if(SelTO.tv_usec < 0) {
            SelTO.tv_usec += 1000000;
            SelTO.tv_sec--;
         }
   
         if(SelTO.tv_sec < 0) {
         // Timeout time has already passed
            SelTO.tv_sec = SelTO.tv_usec = 0;
         }
         else if(SelTO.tv_sec == 0) {
            if(SelTO.tv_usec < 0) {
            // Timeout time has already passed
               SelTO.tv_usec = 0;
            }
         }

#ifdef _WIN32
         // The Windoze control-C handler can't abort the select call so
         // don't wait for more second so we can exit quickly when the 
         // user hits control-C

         if(SelTO.tv_sec > 1) {
            SelTO.tv_sec = 1;
            SelTO.tv_usec = 0;
         }
#endif
         TimeOutSeconds = SelTO.tv_sec;

         if(SelTO.tv_sec == 0 && SelTO.tv_usec == 0) {
            D2PRINTF(("Calling select() %d read, %d write, immediate timeout\n",
                     ReadWaitCnt,WriteWaitCnt));
         }
         else {
            D2PRINTF(("Calling select() %d read, %d write, %ld sec timeout\n",
                     ReadWaitCnt,WriteWaitCnt,TimeOutSeconds));
         }
         Selret = select(FD_SETSIZE,&ReadFDset,&WriteFDset,&ExceptFDset,&SelTO);
      }
      else {
         D2PRINTF(("Calling select() %d read, %d write, no timeout\n",
                  ReadWaitCnt,WriteWaitCnt));
         Selret = select(FD_SETSIZE,&ReadFDset,&WriteFDset,&ExceptFDset,NULL);
      }

#ifndef _WIN32
      if(Selret == SOCKET_ERROR) {
         if(errno != EINTR) {
            LOG_ERROR(("select() failed, %m"));
            Err = ERROR_CODE;
            break;
         }

         if(!bRunning) {
            break;
         }
         else {
            continue;
         }
      }
#else 
      if(Selret == SOCKET_ERROR) {
         Err = ERROR_CODE;
         LOG_ERROR(("select() failed, error = %d",Err));
         break;
      }
#endif

      GetTimeNow();

      pClient1 = (ClientInfo *) avl_t_first(&avl_trans,ClientTree);

      while((pClient = pClient1) != NULL) {
      // NB: Get the next client now in case the current client is deleted
      // when it's processed.
         pClient1 = (ClientInfo *) avl_t_next(&avl_trans);

         if(pClient->bNew) {
         // Ignore this guy, he was created since the last time we called select.
            continue;
         }

         if(pClient->bReadWait && FD_ISSET(pClient->Socket,&ReadFDset)) {
            pClient->bReadReady = TRUE;
            D2PRINTF(("client %u socket ready for read\n",pClient->SN));
         }
         
         if(pClient->bWriteWait && FD_ISSET(pClient->Socket,&WriteFDset)) {
            pClient->bWriteReady = TRUE;
            D2PRINTF(("client %u socket ready for write\n",pClient->SN));
         }

         if(pClient->bTimeWait && 
            (TimeNow.tv_sec > pClient->Timeout.tv_sec ||
            (TimeNow.tv_sec == pClient->Timeout.tv_sec &&
             TimeNow.tv_usec >= pClient->Timeout.tv_usec)) )
         {
               pClient->bTimeOut = TRUE;
               D2PRINTF(("client %u timer timeout\n",pClient->SN));
         }
         
         if(pClient->bReadReady || pClient->bWriteReady || pClient->bTimeOut) {
            CallCurrentState(pClient);
         }
      }
   }
   return Err;
}


int DeleteClient(ClientInfo *pClient)
{
   if(pClient->Socket != INVALID_SOCKET) {
      CLOSE(pClient->Socket);
   }
   if(avl_delete(ClientTree,pClient) == NULL) {
   // This should never happen, log it and core dump
      LOG_ERROR(("DeleteClient(): avl_delete() failed to client %d\n",
                 pClient->SN));
      DumpMe();
   }
   else {
      if(pClient->Buf != NULL) {
         free(pClient->Buf);
      }

      if(pClient->Cleanup != NULL) {
         pClient->Cleanup(pClient);
      }

      free(pClient);
   }

   return FALSE;
}


int SendBuffer(ClientInfo *pClient)
{
   int SendLen;
   int BytesSent;
   int Ret = 0;

   SET_WAIT4_WR(pClient);

   while((SendLen = pClient->BufSize - pClient->Count) > 0) {
      if(SendLen > MAX_SEND_SIZE) {
         SendLen = MAX_SEND_SIZE;
      }
      D3PRINTF(("SendBuffer(), sending %d of %d bytes.\n",SendLen,
               pClient->BufSize));

      BytesSent = WRITE(pClient->Socket,&pClient->Buf[pClient->Count],SendLen);
      if(BytesSent == SOCKET_ERROR ) {
         if(ERROR_CODE == ERROR_WOULD_BLOCK) {
         // buffer full, try again later
            break;
         }
         else {
         // Some kind of error other than buffer full
            DPRINTF(("SendBuffer(), WRITE() returned error %d.\n",ERROR_CODE));
            DeleteClient(pClient);
            break;
         }
      }
      else {
         pClient->Count += SendLen;
      }
   }

   if(SendLen == 0) {
   // We're done with the guy for now, close the socket
      D2PRINTF(("SendBuffer(), send complete.\n"));
      pClient->State = pClient->NextState;
      Ret = 1;
   }

   return Ret;
}


// Read until we get a charcter in the list of Terminator or 
// the buffer fills.
int ReadLine(ClientInfo *pClient)
{
   char Data;
   int  BytesRead;
   int  Ret = 0;

   for( ; ; ) {
      if((BytesRead = READ(pClient->Socket,&Data,1)) == 1) {
         if(Data == pClient->Terminator) {
         // Found the termination character
            pClient->Buf[pClient->Count] = 0;
            D3PRINTF(("ReadLine(): returning \"%s\" to client %u.\n",
                     pClient->Buf,pClient->SN));
            Ret = 1;
            break;
         }
         else {
            pClient->Buf[pClient->Count++] = Data;
         }
      }
      else if(BytesRead == 0) {
      // Connection has closed
         D2PRINTF(("ReadLine(): Client %u connection closed.\n",pClient->SN));
         pClient->Err = ERR_CONNECTION_CLOSED;
         Ret = 1;
         break;
      }
      else if(ERROR_CODE == ERROR_WOULD_BLOCK) {
      // Nothing available to read now, keep waiting
         SET_WAIT4_RD(pClient);
         break;
      }
      else {
      // Some kind of error other than no data available
         DPRINTF(("ReadLine(): READ() returned error %d for client %u.\n",
                  ERROR_CODE,pClient->SN));
         pClient->Err = ERROR_CODE;
         Ret = 1;
      }

      if(pClient->Count == pClient->BufSize) {
      // Buffer overflow !
         LOG_ERROR(("ReadLine(): buffer overflow for client %u.\n",pClient->SN));
         DeleteClient(pClient);
         break;
      }
   }

   if(Ret) {
      pClient->State = pClient->NextState;
   }
   return Ret;
}

// Client compares by IP address: Port
int ClientCompare(const void *avl_a,const void *avl_b,void *avl_param)
{
   ClientInfo *p_a = (ClientInfo *) avl_a;
   ClientInfo *p_b = (ClientInfo *) avl_b;

   if(p_a->HisAdr.i.sin_addr.s_addr == p_b->HisAdr.i.sin_addr.s_addr) {
      return p_a->HisAdr.i.sin_port - p_b->HisAdr.i.sin_port;
   }

   if(p_a->HisAdr.ADDR == p_b->HisAdr.ADDR) {
      return 0;
   }
   if(p_a->HisAdr.ADDR > p_b->HisAdr.ADDR) {
      return 1;
   }
   else {
      return -1;
   }
}


ClientInfo *CreateNewClient()
{
   ClientInfo *pClient = malloc(sizeof(ClientInfo));

   if(pClient != NULL) {
      memset(pClient,0,sizeof(ClientInfo));
      pClient->SN = SN++;
      pClient->Socket = INVALID_SOCKET;
      pClient->bNew = TRUE;
   }
   return pClient;
}


void CallCurrentState(ClientInfo *pClient)
{
   int bLoop = FALSE;

   do {
      if(pClient->State == NULL) {
         LOG_ERROR(("CallCurrentState(): Client %d attempted to call "
                    "undefined state!\n",pClient->SN));
         bLoop = FALSE;
      }
      else {
         bLoop = pClient->State(pClient);
      }
   } while(bLoop);
}

// Set a timeout for <Timeout> milliseconds from now
void SetTimeout(ClientInfo *p,int Timeout)
{
   long microseconds;

   GetTimeNow();
   microseconds = TimeNow.tv_usec + ((Timeout % 1000) * 1000);

   p->bTimeWait = TRUE;
   p->Timeout.tv_usec = microseconds % 1000000;
   p->Timeout.tv_sec  = TimeNow.tv_sec + (Timeout / 1000) + 
                        microseconds / 1000000;
}


void GetTimeNow()
{
#ifdef HAVE_GETTIMEOFDAY
   struct timezone tz;
   gettimeofday(&TimeNow,&tz);
#else
   #ifdef HAVE_FTIME
      // Use ftime() and convert
      #ifndef _WIN32
         struct timeb FTimeNow;
         ftime(&FTimeNow);
      #else
         struct _timeb FTimeNow;
         _ftime(&FTimeNow);
      #endif
      
      TimeNow.tv_sec = FTimeNow.time;
      TimeNow.tv_usec = FTimeNow.millitm * 1000;
   #else
   // How the heck do you get the time of day on this OS ?
      #error Need either ftime() or gettimeofday() !
   #endif
#endif
}

void OpComplete(ClientInfo *pClient, int Err)
{
   pClient->Err = Err;
   if(pClient->Complete != NULL) {
      pClient->Complete(pClient);
   }
   else {
      DeleteClient(pClient);
   }
}


void BufPrintf(ClientInfo *p,char *fmt,...)
{
   va_list args;
   
   va_start(args,fmt);
   
   if(p->Count < p->BufSize) {
      p->Count += vsnprintf(&p->Buf[p->Count],p->BufSize-p->Count,fmt,args);
   }
   va_end(args);
}


void LogNorm(char *fmt, ...)
{
   va_list args;
   va_start(args,fmt);

#ifndef _WIN32
   if(bDaemon) {
      vsyslog(LOG_INFO,fmt,args);
   }

   if(!bDaemon || DebugLevel > 0) {
      vprintf(fmt,args);
      printf("\n");
   }
#else
   vprintf(fmt,args);
   printf("\n");
#endif
   va_end(args);
}

void LogWarn(char *fmt, ...)
{
   va_list args;
   va_start(args,fmt);

#ifndef _WIN32
   if(bDaemon) {
      vsyslog(LOG_WARNING,fmt,args);
   }
   
   if(!bDaemon || DebugLevel > 0) {
      vprintf(fmt,args);
      printf("\n");
   }
#else
   vprintf(fmt,args);
   printf("\n");
#endif
   va_end(args);
}

void LogErr(char *fmt, ...)
{
   va_list args;
   va_start(args,fmt);

#ifndef _WIN32
   if(bDaemon) {
      vsyslog(LOG_ERR,fmt,args);
   }

   if(!bDaemon || DebugLevel > 0) {
      vprintf(fmt,args);
      printf("\n");
   }
#else
   vprintf(fmt,args);
   printf("\n");
#endif
   va_end(args);
}

/* Force a code dump */
void DumpMe()
{
   char *cp = NULL;

   *cp = 0;
}

#ifndef _WIN32
void SigHandler(int Signal)
{
   switch(Signal) {
      case SIGHUP:
         LOG_NORM(("Received SIGHUP, reloading conf file"));
         bRefreshConfig = TRUE;
         break;

      case SIGINT:
         LOG_NORM(("Received SIGINT, shutting down"));
         bShutdown = TRUE;
         break;

      case SIGTERM:
         LOG_NORM(("Received SIGTERM, shutting down"));
         bShutdown = TRUE;
         break;
   }
}

#else // Windoze
   
BOOL SigHandler(DWORD Type) 
{ 
   if(Type == CTRL_C_EVENT) {
   // Handle the CTRL+C signal. 
      LOG_NORM(("Received SIGINT, shutting down"));
      bShutdown = TRUE;
      return TRUE; 
   } 
   return FALSE; 
} 
 
#endif
